from typing import Dict

import numpy as np
import pybullet as p

from pybullet_industrial.endeffector_tool import EndeffectorTool
from pybullet_industrial.robot_base import RobotBase


class Camera(EndeffectorTool):
    """Special Endeffector Tool which can cast rays. Base for Extruder and Remover classes.

    Args:
        urdf_model (str): A valid path to a urdf file describint the tool geometry
        start_position (np.array): the position at which the tool should be spawned
        start_orientation (np.array): the orientation at which the tool should be spawned
        camera_parameters (Dict[str,float]): A dictionary containing the camera parameters.
                                             Default Values are:
                                             'width': 480,
                                             'height': 240,
                                             'fov': 60,
                                             'aspect ratio': 1,
                                             'near plane distance': 0.01,
                                             'far plane distance': 100
        coupled_robot (RobotBase, optional): A pybullet_industrial.RobotBase object if
                                             the robot is coupled from the start.
                                             Defaults to None.
        camera_frame (str, optional): The name of the urdf_link
                                      at which the camera is located.
                                      Defaults to None in which case the last link is used.
        connector_frame (str, optional): The name of the urdf_link
                                         at which a robot connects.
                                         Defaults to None in which case the base link is used.
    """

    def __init__(self, urdf_model: str, start_position: np.array, start_orientation: np.array,
                 camera_parameters: Dict, coupled_robot: RobotBase = None,
                 camera_frame: str = None, connector_frame: str = None):

        super().__init__(urdf_model, start_position, start_orientation,
                         coupled_robot, camera_frame, connector_frame)

        self.camera_parameters = {'width': 480,
                                  'height': 240,
                                  'fov': 60,
                                  'aspect ratio': 1,
                                  'near plane distance': 0.01,
                                  'far plane distance': 100}
        self.set_camera_parameters(camera_parameters)

        self.projection_matrix = self.__get_projection_matrix()

    def __get_projection_matrix(self):
        fov = self.camera_parameters['fov']
        aspect_ratio = self.camera_parameters['aspect ratio']
        near_plane = self.camera_parameters['near plane distance']
        far_plane = self.camera_parameters['far plane distance']
        projection_matrix = p.computeProjectionMatrixFOV(
            fov, aspect_ratio, near_plane, far_plane)
        return projection_matrix

    def set_camera_parameters(self, camera_parameters: Dict):
        """Sets the camera parameters

        Args:
            camera_parameters (Dict[str,float]): A dictionary containing the camera parameters
        Raises:
            KeyError: If a key is not a valid parameter
        """
        for key in camera_parameters:
            if not key in self.camera_parameters:
                raise KeyError("The specified property keys are not valid" +
                               " Valid keys are: "+str(self.camera_parameters))
            self.camera_parameters[key] = camera_parameters[key]

        self.projection_matrix = self.__get_projection_matrix()

    def get_image(self):
        """Captures a camera image of the current simulation

        Returns:
            np.array: A array of pixel colors in rgba format in 255 color format
        """
        link_state = p.getLinkState(
            self.urdf, self._tcp_id, computeForwardKinematics=True)
        com_p = np.array(link_state[0])
        com_o = np.array(link_state[1])
        rot_matrix = p.getMatrixFromQuaternion(com_o)
        rot_matrix = np.array(rot_matrix).reshape(3, 3)
        # Initial vectors
        init_camera_vector = (0, 0, 1)  # z-axis
        init_up_vector = (0, 1, 0)  # y-axis
        # Rotated vectors
        camera_vector = rot_matrix.dot(init_camera_vector)
        up_vector = rot_matrix.dot(init_up_vector)
        view_matrix = p.computeViewMatrix(
            com_p, com_p + 0.1 * camera_vector, up_vector)

        width = self.camera_parameters['width']
        height = self.camera_parameters['height']
        images = p.getCameraImage(
            width, height, view_matrix, self.projection_matrix)
        img = np.reshape(images[2], (height, width, 4))
        return img



from typing import Dict

import numpy as np
import pybullet as p

from pybullet_industrial.raycaster import RayCaster
from pybullet_industrial.robot_base import RobotBase


class Remover(RayCaster):
    """Special Remover Tool which can remove objects from the simulation.

    Args:
        urdf_model (str): A valid path to a urdf file describint the tool geometry
        start_position (np.array): the position at which the tool should be spawned
        start_orientation (np.array): the orientation at which the tool should be spawned
        remover_properties(Dict): A dictionary containing the properties of the extrusion head.
                                  During initialization only 'material' has to be set.
                                  Default Values are:
                                  'opening angle':0,'number of rays':1,
                                  'maximum distance':1
        coupled_robot (RobotBase, optional): A pybullet_industrial.RobotBase object if
                                             the robot is coupled from the start.
                                             Defaults to None.
        tcp_frame (str, optional): The name of the urdf_link
                                   describing the tool center point.
                                   Defaults to None in which case the last link is used.
        connector_frame (str, optional): The name of the urdf_link
                                         at which a robot connects.
                                         Defaults to None in which case the base link is used.
    """

    def __init__(self, urdf_model: str, start_position: np.array, start_orientation: np.array,
                 remover_properties: Dict, coupled_robot: RobotBase = None,
                 tcp_frame: str = None, connector_frame: str = None):

        super().__init__(urdf_model, start_position, start_orientation,
                         coupled_robot, tcp_frame, connector_frame)

        self.change_properties(remover_properties)

    def calculate_process_force(self, ray_cast_results, tcp_frame=None):
        """Function calculating the process force of the milling tool.

        Args:
            ray_cast_results (list): The result of a batch ray cast
                                       as performed by the remove function
            tcp_frame (str): The name of the tool center point frame. Defaults to None.

        Returns:
            np.array: a 3 dimensional array containing the process force in the x,y,z direction
        """

        return np.zeros(3)

    def remove(self, tcp_frame=None):
        """Removes simulation elements from the simulation environment.

        Args:
            tcp_frame (str, optional): the name of the link from which to remove the material.
                                       Defaults to None in which case the default tcp is used
        """
        position, orientation = self.get_tool_pose(tcp_frame)
        ray_cast_results = self.cast_rays(position, orientation)

        process_force = self.calculate_process_force(
            ray_cast_results, tcp_frame)
        self.apply_tcp_force(process_force, tcp_frame)

        removed_objects = []
        for ray_intersection in ray_cast_results:
            if ray_intersection[0] != -1:
                p.removeBody(ray_intersection[0])
                removed_objects.append(ray_intersection[0])
        return removed_objects



# setup.py
import os

import setuptools

with open("requirements.txt", "r") as fh:
    requirements = fh.readlines()

setup_py_dir = os.path.dirname(os.path.realpath(__file__))

need_files = []
datadir = "triped_sim"

hh = setup_py_dir + "/" + datadir

for root, dirs, files in os.walk(hh):
    for fn in files:
        ext = os.path.splitext(fn)[1][1:]
        if ext and ext in 'png gif jpg urdf sdf obj mtl dae off stl STL xml '.split():
            fn = root + "/" + fn
            need_files.append(fn[1+len(hh):])

setuptools.setup(
    name='pybullet_industrial',
    version='0.9',
    author='Jan Baumgärtner',
    description='Pybullet_industrial is a process-aware robot simulation framework for pybullet.',
    install_requires=[req for req in requirements if req[:2] != "# "],
    packages=setuptools.find_packages(),
    package_data={'pybullet_industrial': need_files}
)



from typing import Dict

import numpy as np
import pybullet as p

from pybullet_industrial.endeffector_tool import EndeffectorTool
from pybullet_industrial.robot_base import RobotBase


class RayCaster(EndeffectorTool):
    """Special Endeffector Tool which can cast rays. Base for Extruder and Remover classes.

    Args:
        urdf_model (str): A valid path to a urdf file describint the tool geometry
        start_position (np.array): the position at which the tool should be spawned
        start_orientation (np.array): the orientation at which the tool should be spawned
        raycast_properties(Dict): A dictionary containing the properties of the extrusion head.
                                  During initialization only 'material' has to be set.
                                  Default Values are:
                                  'opening angle':0,'number of rays':1,
                                  'maximum distance':1
        coupled_robot (RobotBase, optional): A pybullet_industrial.RobotBase object if
                                             the robot is coupled from the start.
                                             Defaults to None.
        tcp_frame (str, optional): The name of the urdf_link
                                   describing the tool center point.
                                   Defaults to None in which case the last link is used.
        connector_frame (str, optional): The name of the urdf_link
                                         at which a robot connects.
                                         Defaults to None in which case the base link is used.
        """

    def __init__(self, urdf_model: str, start_position: np.array, start_orientation: np.array,
                 raycast_properties: Dict, coupled_robot: RobotBase = None,
                 tcp_frame: str = None, connector_frame: str = None):

        super().__init__(urdf_model, start_position, start_orientation,
                         coupled_robot, tcp_frame, connector_frame)

        self.properties = {'opening angle': 0,
                           'number of rays': 1,
                           'maximum distance': 1}
        if raycast_properties is not None:
            self.change_properties(raycast_properties)

    def change_properties(self, new_properties: Dict):
        """Allows retroactive changes to the ray casting properties.


        Args:
            new_properties (Dict): A dictionary containing values and keys that
                                   should be changed

        Raises:
            KeyError: If a key is not a valid property
        """
        for key in new_properties:
            if not key in self.properties:
                raise KeyError("The specified property keys are not valid" +
                               " Valid keys are: "+str(self.properties.keys()))
            self.properties[key] = new_properties[key]

    def cast_rays(self, position: np.array, orientation: np.array):
        """Randomly casts rays withn a given range from a specified position and orientation

        Args:
            position (np.array): start position of the raycast
            orientation (np.array): start orientation of the raycast

        Returns:
            List: The result of a raycast
        """
        opening_angle = self.properties['opening angle']
        number_of_rays = self.properties['number of rays']
        ray_length = self.properties['maximum distance']
        ray_start_pos = []
        ray_end_pos = []

        phi = np.random.uniform(-np.pi, np.pi, number_of_rays)
        theta = np.random.uniform(-0.5*opening_angle,
                                  0.5*opening_angle, number_of_rays)
        x = np.sin(theta) * np.cos(phi)
        y = np.sin(theta) * np.sin(phi)
        z = np.cos(theta)
        ray_directions = np.array([x, y, z])

        rot_matrix = p.getMatrixFromQuaternion(orientation)
        rot_matrix = np.array(rot_matrix).reshape(3, 3)

        for i in range(number_of_rays):
            ray_start_pos.append(position)

            ray_dir = rot_matrix@ray_directions[:, i]

            ray_end_pos.append(position-ray_length*ray_dir)

        results = p.rayTestBatch(ray_start_pos, ray_end_pos)
        return results
